/*****************************************************************************/
/*!
 * \file expr_op.cpp
 * 
 * Author: Sergey Berezin
 * 
 * Created: Fri Feb  7 15:29:42 2003
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/

#include "expr_op.h"
#include "expr.h"
#include "expr_manager.h"
#include "compat_hash_set.h"
#include <iostream>
#include <sstream>

using namespace std;

namespace CVCL {

  // Construct an operator from a kind.  If ExprManager is provided,
  // Op will be able to print itself nicely.
//   Op::Op(int kind) : d_kind(kind) {
//     d_em = NULL;
//     // Since no ExprManager is given, we cannot test if the kind is
//     // registered
//     DebugAssert(!hasExpr(),
// 		"CVCL::Op::Op(kind): "
// 		"must supply the operator expression");
//   }

  Op::Op(ExprManager *em, int kind) : d_expr(em, kind) // d_kind(kind)
  {
//     DebugAssert(em->isKindRegistered(kind),
//  		("CVCL::Op::Op(kind = " + int2string(kind)
//   		 + "): kind is not registered").c_str());
//     d_em = em;
//     DebugAssert(!hasExpr(),
// 		"CVCL::Op::Op(kind): "
// 		"must supply the operator expression");
  }

  Op::Op(const Expr& e): d_expr(e) {
//     DebugAssert(e.getEM()->isKindRegistered(kind),
// 		("CVCL::Op::Op(kind = " + int2string(kind)
//   		 + "): kind is not registered").c_str());
//     DebugAssert(hasExpr(),
// 		("CVCL::Op::Op(kind = " + e.getEM()->getKindName(kind)
// 		 + ", Expr = " + e.toString() + "): wrong kind").c_str());
//     d_expr = e;
//     d_em = e.getEM();
  }




  // Copy constructor
  Op::Op(const Op& op): // d_kind(op.d_kind), 
    d_expr(op.d_expr) // , d_em(op.d_em)
 { }

  // A constructor that rebuilds the Op for the given ExprManager
  Op::Op(ExprManager* em, const Op& op) // : d_kind(op.d_kind)
  {
//     DebugAssert(em->isKindRegistered(op.d_kind),
// 		"CVCL::Op::Op(em, Op(kind = " + int2string(op.d_kind)
// 		+ ")): kind is not registered");
//     if(op.hasExpr()) 
    d_expr = em->rebuildExpr(op.getExpr());
    // d_em = em;
  }

  // Assignment operator
  Op& Op::operator=(const Op& op) {
    if(&op == this) return *this; // Self-assignment
    // d_kind = op.d_kind;
    d_expr = op.d_expr;
    // d_em = op.d_em;
    return *this;
  }

  // Check if op has an associated Expr
  /*
  bool Op::hasExpr() const {
    switch(d_kind) {
    case UCONST:
    case LAMBDA:
    case FORALL:
    case EXISTS:
    case RECORD:
    case RECORD_TYPE:
    case RECORD_SELECT:
    case RECORD_UPDATE:
    case TUPLE_SELECT:
    case TUPLE_UPDATE:
      return true;
    default:
      return (!d_expr.isNull() || hasExprKind(d_kind));
    }
  }
  */


  string Op::toString() const {
    ostringstream ss;
    ss << *this;
    /*
    ss << "Op(";
    if(d_em != NULL) ss << d_em->getKindName(d_kind);
    else
      ss << "kind = " << d_kind;
    if(hasExpr()) ss << ", " << d_expr;
    ss << ")";
    */
    return ss.str();
  }

  /*
  bool operator==(const Op& op1, const Op& op2) {
    if(op1.getKind() != op2.getKind()) return false;
    if(op1.hasExpr()) {
      return op1.getExpr() == op2.getExpr();
    }
    else return true;
  }

  Op getOp(const Expr& e) {
    DebugAssert(!e.isNull(), "Op() is called on Null expression");
    int kind = e.getKind();
    Op op(NULL_KIND);
    //    if(hasExprKind(kind))
    if(e.hasOp()) op = Op(kind, e.getOp());
    else if(e.isClosure()) op = Op(kind, e);
    else
      switch(kind) {
      case RATIONAL_EXPR:
	op = Op(kind, e);
	break;
      case RECORD:
      case RECORD_TYPE:
	op = Op(kind, newExpr(e.getEM(), kind, e.getRecordFields(),
			      vector<Expr>()));
	break;
      case RECORD_SELECT:
      case RECORD_UPDATE:
	op = Op(kind, newExpr(e.getEM(), kind, e.getRecordField(),
			      vector<Expr>()));
	break;
      case TUPLE_SELECT:
      case TUPLE_UPDATE:
	op = Op(kind, newExpr(e.getEM(), kind, e.getTupleIndex(),
			      vector<Expr>()));
	break;
      case UCONST:
	if(e.lookupType().isFunction())
	  op = Op(UFUNC, e);
	else
	  op = Op(kind, e);
	break;
      default:
	op = Op(e.getEM(), kind);
	break;
      }
    TRACE("op", "getOp("+e.toString()+") = ", op, "");
    return op;
  }

  */
} // end of namespace CVCL
